#ifndef __CCU_BSP_H__
#define __CCU_BSP_H__

void ccu_Set_LED(void *self, int LEDID, int on);
void ccu_Set_FanSSR(void *self, int on);
void ccu_Set_HPumpSSR(void *self, int on);
int ccu_Get_FanFBA(void *self);
int ccu_Get_FanFBB(void *self);
int ccu_Get_FanFBC(void *self);
int ccu_Get_WATFB(void *self);
int ccu_Get_GATFB(void *self);
int ccu_Get_EMGFB(void *self);
int ccu_Get_LGPTFA(void *self);
int ccu_Get_LGPTFB(void *self);
float ccu_Get_BOX_Temp(void *self, int idx);
float ccu_Get_GUN_Temp(void *self, int idx);
int ccu_Get_NFC_CardID(void* self, int timeout, char* card_id);

void ccu_Read_DCMeterA(void *self, float* VMeter, float* IMeter, float* PMeter, float* EMeter);
void ccu_Read_DCMeterB(void *self, float* VMeter, float* IMeter, float* PMeter, float* EMeter);
void ccu_set_DCGroundA_state(void* self, int onoff);
void ccu_set_DCGroundB_state(void* self, int onoff);
void ccu_Read_DCGroundA(void *self, float* VGround, float* RGround);
void ccu_Read_DCGroundB(void *self, float* VGround, float* RGround);

int ccu_BSP_init(void);
void ccu_BSP_exit(void);

#endif//__CCU_BSP_H__
